@@ -39,6 +39,7 @@
 #include <QFormLayout>
 #include <QDialogButtonBox>
 
+#include "mprocess.h"
 #include "msettings.h"
 #include "autopart.h"
 #include "partman.h"
@@ -47,16 +48,15 @@ PartMan::PartMan(MProcess &mproc, Ui::MeInstall &ui, const QSettings &appConf, c
     : QAbstractItemModel(ui.boxMain), proc(mproc), root(DeviceItem::Unknown), gui(ui)
 {
     const QString &projShort = appConf.value("PROJECT_SHORTNAME").toString();
-    defaultLabels["ESP"] = "EFI System";
-    defaultLabels["BIOS-GRUB"] = "BIOS GRUB";
-    defaultLabels["/boot"] = "boot";
-    defaultLabels["/"] = "root" + projShort + appConf.value("VERSION").toString();
-    defaultLabels["/home"] = "home" + projShort;
-    defaultLabels["SWAP"] = "swap" + projShort;
+    volSpecs["ESP"] = {"EFI System"};
+    volSpecs["BIOS-GRUB"] = {"BIOS GRUB"};
+    volSpecs["/boot"] = {"boot"};
+    volSpecs["/"] = {"root" + projShort + appConf.value("VERSION").toString()};
+    volSpecs["/home"] = {"home" + projShort};
+    volSpecs["SWAP"] = {"swap" + projShort};
     brave = appArgs.isSet("brave");
     gptoverride = appArgs.isSet("gpt-override");
 
-    master = gui.boxMain;
     root.partman = this;
     gui.treePartitions->setModel(this);
     gui.treePartitions->setItemDelegate(new DeviceItemDelegate);
@@ -503,20 +503,23 @@ void PartMan::treeMenu(const QPoint &)
     } else if (twit->type == DeviceItem::Drive) {
         QAction *actAdd = menu.addAction(tr("&Add partition"));
         actAdd->setEnabled(gui.pushPartAdd->isEnabled());
-        QAction *actBuild = menu.addAction(tr("&Build layout..."));
         menu.addSeparator();
         QAction *actClear = menu.addAction(tr("New &layout"));
         QAction *actReset = menu.addAction(tr("&Reset layout"));
+        menu.addSeparator();
+        QAction *actBuilder = menu.addAction(tr("Layout &Builder..."));
 
         const bool locked = twit->isLocked();
         actClear->setDisabled(locked);
         actReset->setDisabled(locked);
-        actBuild->setDisabled(locked);
+
+        const long long minSpace = brave ? 0 : volSpecTotal("/", QStringList()).minimum;
+        actBuilder->setEnabled(!locked && autopart && twit->size >= minSpace);
 
         QAction *action = menu.exec(QCursor::pos());
         if (action == actAdd) partAddClick(true);
         else if (action == actClear) partClearClick(true);
-        else if (action == actBuild) {
+        else if (action == actBuilder) {
             if (autopart) autopart->builderGUI(twit);
             treeSelChange();
         } else if (action == actReset) {
@@ -592,7 +595,7 @@ void PartMan::partManRunClick()
 // Partition menu items
 void PartMan::partMenuUnlock(DeviceItem *twit)
 {
-    QDialog dialog(master);
+    QDialog dialog(gui.boxMain);
     QFormLayout layout(&dialog);
     dialog.setWindowTitle(tr("Unlock Drive"));
     QLineEdit *editVDev = new QLineEdit(&dialog);
@@ -629,7 +632,7 @@ void PartMan::partMenuUnlock(DeviceItem *twit)
             qDebug() << "partMenuUnlock() failed:" << msg;
             gui.boxMain->setEnabled(true);
             qApp->restoreOverrideCursor();
-            QMessageBox::warning(master, QString(),
+            QMessageBox::warning(gui.boxMain, QString(),
                 tr("Could not unlock device. Possible incorrect password."));
         }
     }
@@ -659,7 +662,7 @@ void PartMan::partMenuLock(DeviceItem *twit)
     qApp->restoreOverrideCursor();
     // If not OK then a command failed, or trying to close a non-LUKS device.
     if (!ok) {
-        QMessageBox::critical(master, QString(), tr("Failed to close %1").arg(twit->device));
+        QMessageBox::critical(gui.boxMain, QString(), tr("Failed to close %1").arg(twit->device));
     }
 }
 
@@ -703,7 +706,7 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
             if (cmptext.count(QRegularExpression("[^A-Z0-9\\/\\@\\.\\-\\_]|\\/\\/"))) ok = false;
             if (cmptext.startsWith('/') || cmptext.endsWith('/')) ok = false;
             if (!ok) {
-                QMessageBox::critical(master, QString(), tr("Invalid subvolume label"));
+                QMessageBox::critical(gui.boxMain, QString(), tr("Invalid subvolume label"));
                 return false;
             }
             // Check for duplicate subvolume label entries.
@@ -714,7 +717,7 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
             for (int ixi = 0; ixi < count; ++ixi) {
                 if (ixi == index) continue;
                 if (pit->child(ixi)->label.trimmed().toUpper() == cmptext) {
-                    QMessageBox::critical(master, QString(), tr("Duplicate subvolume label"));
+                    QMessageBox::critical(gui.boxMain, QString(), tr("Duplicate subvolume label"));
                     return false;
                 }
             }
@@ -722,7 +725,7 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
         QString mount = item->realUseFor();
         if (mount.isEmpty()) continue;
         if (!mount.startsWith("/") && !item->allowedUsesFor().contains(mount)) {
-            QMessageBox::critical(master, QString(),
+            QMessageBox::critical(gui.boxMain, QString(),
                 tr("Invalid use for %1: %2").arg(item->shownDevice(), mount));
             return false;
         }
@@ -734,7 +737,7 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
 
         // The mount can only be selected once.
         if (twit) {
-            QMessageBox::critical(master, QString(), tr("%1 is already"
+            QMessageBox::critical(gui.boxMain, QString(), tr("%1 is already"
                 " selected for: %2").arg(twit->shownDevice(), twit->shownUseFor()));
             return false;
         } else if(!mount.isEmpty() && mount != "FORMAT" && mount != "ESP" && mount != "BIOS-GRUB"){
@@ -755,23 +758,25 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
     }
 
     DeviceItem *rootitem = mounts.value("/");
-    if (!rootitem || rootitem->size < rootSpaceNeeded) {
-        const QString &tmin = QLocale::system().formattedDataSize(rootSpaceNeeded + 1048575,
-            1, QLocale::DataSizeTraditionalFormat);
-        QMessageBox::critical(master, QString(),
-            tr("A root partition of at least %1 is required.").arg(tmin));
+    DeviceItem *bootitem = mounts.value("/boot");
+    const long long rootMin = volSpecTotal("/").minimum;
+    const QString &tMinRoot = QLocale::system().formattedDataSize(rootMin,
+        1, QLocale::DataSizeTraditionalFormat);
+    if (!rootitem) {
+        QMessageBox::critical(gui.boxMain, QString(),
+            tr("A root partition of at least %1 is required.").arg(tMinRoot));
         return false;
     } else {
         if (!rootitem->willFormat() && mounts.contains("/home")) {
             const QString errmsg = tr("Cannot preserve /home inside root (/) if"
                 " a separate /home partition is also mounted.");
-            QMessageBox::critical(master, QString(), errmsg);
+            QMessageBox::critical(gui.boxMain, QString(), errmsg);
             return false;
         }
         if (rootitem->encrypt) encryptRoot = true;
     }
-    if (encryptRoot && !mounts.contains("/boot")) {
-        QMessageBox::critical(master, QString(),
+    if (encryptRoot && !bootitem) {
+        QMessageBox::critical(gui.boxMain, QString(),
             tr("You must choose a separate boot partition when encrypting root."));
         return false;
     }
@@ -779,6 +784,8 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
     if (!automatic) {
         // Final warnings before the installer starts.
         QString details, biosgpt;
+        QStringList tooSmall;
+        const QString &msgszItem(" - " + tr("%1 (%2) requires %3"));
         for (int ixdrv = 0; ixdrv < root.childCount(); ++ixdrv) {
             DeviceItem *drvit = root.child(ixdrv);
             const int partCount = drvit->childCount();
@@ -805,6 +812,14 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
                 }
                 details += actmsg.arg(partit->shownDevice(), partit->shownUseFor()) + '\n';
                 if (use == "BIOS-GRUB") hasBiosGrub = true;
+
+                // If this volume is too small, add to the warning list.
+                const long long minSize = volSpecTotal(partit->realUseFor()).minimum;
+                if (partit->size < minSize) {
+                    const QString &mstr = QLocale::system().formattedDataSize(minSize,
+                        1, QLocale::DataSizeTraditionalFormat);
+                    tooSmall.append(msgszItem.arg(partit->shownUseFor(), partit->device, mstr));
+                }
             }
             // Potentially unbootable GPT when on a BIOS-based PC.
             const bool hasBoot = (drvit->active != nullptr);
@@ -813,10 +828,18 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
             }
         }
         // Warning messages
-        QMessageBox msgbox(master);
+        QMessageBox msgbox(gui.boxMain);
         msgbox.setIcon(QMessageBox::Warning);
         msgbox.setStandardButtons(QMessageBox::Yes|QMessageBox::No);
         msgbox.setDefaultButton(QMessageBox::No);
+        if (!tooSmall.isEmpty()) {
+            tooSmall.sort();
+            msgbox.setText(tr("The installation may fail because"
+                " the following volumes are too small:") + "\n\n"
+                + tooSmall.join('\n')
+                + "\n\n" + tr("Are you sure you want to continue?"));
+            if (msgbox.exec() != QMessageBox::Yes) return false;
+        }
         if (!biosgpt.isEmpty()) {
             biosgpt.prepend(tr("The following drives are, or will be, setup with GPT,"
                 " but do not have a BIOS-GRUB partition:") + "\n\n");
@@ -827,7 +850,15 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
         }
         msgbox.setText(tr("The %1 installer will now perform the requested actions.").arg(project));
         msgbox.setInformativeText(tr("These actions cannot be undone. Do you want to continue?"));
+        details.chop(1); // Remove trailing new-line character.
         msgbox.setDetailedText(details);
+        const QList<QAbstractButton *> &btns = msgbox.buttons();
+        for (auto btn : btns) {
+            if (msgbox.buttonRole(btn) == QMessageBox::ActionRole) {
+                btn->click(); // Simulate clicking the "Show Details..." button.
+                break;
+            }
+        }
         if (msgbox.exec() != QMessageBox::Yes) return false;
     }
 
@@ -878,12 +909,12 @@ bool PartMan::checkTargetDrivesOK()
         msg += tr("If unsure, please exit the Installer and run GSmartControl for more information.") + "\n\n";
         if (!smartFail.isEmpty()) {
             msg += tr("Do you want to abort the installation?");
-            ans = QMessageBox::critical(master, QString(), msg,
+            ans = QMessageBox::critical(gui.boxMain, QString(), msg,
                     QMessageBox::Yes|QMessageBox::Default|QMessageBox::Escape, QMessageBox::No);
             if (ans == QMessageBox::Yes) return false;
         } else {
             msg += tr("Do you want to continue?");
-            ans = QMessageBox::warning(master, QString(), msg,
+            ans = QMessageBox::warning(gui.boxMain, QString(), msg,
                     QMessageBox::Yes|QMessageBox::Default, QMessageBox::No|QMessageBox::Escape);
             if (ans != QMessageBox::Yes) return false;
         }
@@ -893,7 +924,7 @@ bool PartMan::checkTargetDrivesOK()
 
 void PartMan::luksFormat(DeviceItem *partit, const QByteArray &password)
 {
-    QStringList cargs({"--batch-mode", "--key-size=512", "--hash=sha512", "--pbkdf=argon2id"});
+    QStringList cargs({"--batch-mode", "--key-size=512", "--hash=sha512"});
     if (partit->physec > 0) cargs.append("--sector-size=" + QString::number(partit->physec));
     cargs << "luksFormat" << partit->path;
     if (!proc.exec("cryptsetup", cargs, &password)) {
@@ -1442,6 +1473,20 @@ DeviceItem *PartMan::findHostDev(const QString &path) const
     return devit;
 }
 
+struct PartMan::VolumeSpec PartMan::volSpecTotal(const QString &path, const QStringList &vols) const
+{
+    struct VolumeSpec vspec = volSpecs.value(path);
+    const QString &spath = (path!='/' ? path+'/' : path);
+    for (const auto &it : volSpecs.toStdMap()) {
+        if (it.first.size() > 1 && it.first.startsWith(spath) && !vols.contains(it.first)) {
+            vspec.image += it.second.image;
+            vspec.minimum += it.second.minimum;
+            vspec.preferred += it.second.preferred;
+        }
+    }
+    return vspec;
+}
+
 /*************************\
  * Model View Controller *
 \*************************/
@@ -1814,9 +1859,8 @@ QString DeviceItem::realUseFor(const QString &use)
     else if (!use.startsWith('/')) return use.toUpper();
     else return use;
 }
-QString DeviceItem::shownUseFor() const
+QString DeviceItem::shownUseFor(const QString &use)
 {
-    const QString &use = realUseFor();
     if (use == "/") return "/ (root)";
     else if (use == "ESP") return tr("EFI System Partition");
     else if (use == "SWAP") return tr("swap space");
@@ -1893,21 +1937,26 @@ QStringList DeviceItem::allowedUsesFor(bool real, bool all) const
 {
     if (!isVolume() && type != Subvolume) return QStringList();
     QStringList list;
-    if (!partman || size >= partman->rootSpaceNeeded) list << "root";
-    if (type == Subvolume) list << "home"; // swap requires Linux 5.0 or later
+    auto checkAndAdd = [&](const QString &use) {
+        const QString &realUse = realUseFor(use);
+        if (all || !partman || size >= partman->volSpecs.value(realUse).minimum) {
+            list.append(real ? realUse : use);
+        }
+    };
+
+    if (type != Subvolume) checkAndAdd("Format");
+    checkAndAdd("root");
+    if (type == Subvolume) checkAndAdd("home"); // swap requires Linux 5.0 or later
     else {
-        list.prepend("Format");
         if (type != VirtualBD) {
-            if (all || size <= 16*MB) list << "BIOS-GRUB";
+            if (all || size <= 16*MB) checkAndAdd("BIOS-GRUB");
             if (all || size <= 8*GB) {
-                if (size < (2*TB - 512)) list << "ESP";
-                list << "boot";
+                if (size < (2*TB - 512)) checkAndAdd("ESP");
+                checkAndAdd("boot");
             }
         }
-        list << "swap" << "home";
-    }
-    if (real) {
-        for(QString &use : list) use = realUseFor(use);
+        checkAndAdd("swap");
+        checkAndAdd("home");
     }
     return list;
 }
@@ -2023,7 +2072,7 @@ void DeviceItem::autoFill(unsigned int changed)
             }
             label = newLabel;
         } else if (partman) {
-            label = partman->defaultLabels.value(use);
+            label = partman->volSpecs.value(use).defaultLabel;
         }
         // Automatic default boot device selection
         if ((type != VirtualBD) && (use == "/boot" || use == "/")) {
